@@ -10,7 +10,7 @@ use LWP::UserAgent;
 use Data::Dumper;
 
 @ISA = qw(LWP::UserAgent);
-$VERSION = '0.09';
+$VERSION = '0.10';
 
 my $UA = __PACKAGE__->new;
 $UA->agent(join "/", __PACKAGE__, $VERSION);
@@ -144,50 +144,70 @@ sub process_text {
 
 	# Pre-process the beginning of document so we can appropriately handled bad documents
   # that are missing the <BODY> tag
-	my ($found_html, @tokens_up_to_html, $found_head, @tokens_up_to_head, $found_body, @tokens);
+	my ($found_html, @tokens_up_to_html, $found_head, @tokens_up_to_head, $found_body, @tokens,
+			$found_close_body, @tokens_before_close_html, $found_close_html);
 	while (my $token = $parser->get_token) {
-		# If it's an open html, save the previous tokens and this one in a seperate array
+		# If it's <html> save the previous tokens and this one in a seperate array
 		if ($token->[0] eq 'S' and $token->[1] eq 'html') {
 			$found_html++;
 			push @tokens_up_to_html, @tokens, $token;
 			$r->log->debug("HTML Tokens: ", Dumper(@tokens_up_to_html), qq(\n\n));
 			@tokens = ();
 		} # End if
-		# If it's a close head, save the previous tokens and this one in a seperate array
+		# If it's </head> save the previous tokens and this one in a seperate array
 		elsif ($token->[0] eq 'E' and $token->[1] eq 'head') {
 			$found_head++;
 			push @tokens_up_to_head, @tokens, $token;
 			$r->log->debug("</HEAD> Tokens: ", Dumper(@tokens_up_to_head), qq(\n\n));
 			@tokens = ();
-		} # End if
-		# If it's an open body, add this one and the remaining to the tokens we've seen and quit
+		} # End elsif
+		# If it's <body> add this one to the stack and set a flag
 		elsif ($token->[0] eq 'S' and $token->[1] eq 'body') {
 			$found_body++;
 			push @tokens, $token;
-			push @tokens, $token while $token = $parser->get_token;
-			last;
-		} # # End if
+		} # End elsif
+		# If it's </BODY> add this one to the stack and set a flag
+		elsif ($token->[0] eq 'E' and $token->[1] eq 'body') {
+			$found_close_body++;
+			push @tokens, $token;
+		} # End elsif
+		# If it's </HTML> save the previous tokens in a seperate array
+		elsif ($token->[0] eq 'E' and $token->[1] eq 'html') {
+			$found_close_html++;
+			push @tokens_before_close_html, @tokens;
+			@tokens = $token;
+		} # End elsif
 		# Otherwise just save up the tokens
 		else {push @tokens, $token}
 	} # End while
 
-	# Build our body tag in case we need it
+	# Build our body tags in case we need them
 	my $body_tag = ['S', 'body', {}, [], '<body>'];
+	my $close_body_tag = ['E', 'body', {}, [], '</body>'];
 
 	# Rebuild the master array of tokens
 	# If we found <body> just make one big of array of the tokens we saved
 	my @all_tokens;
-	if ($found_body) {@all_tokens = (@tokens_up_to_html, @tokens_up_to_head, @tokens)}
-	
+	if ($found_body) {@all_tokens = (@tokens_up_to_html, @tokens_up_to_head)}
+
 	# If we found </head> but no <body> add <body> after the </head>
-	elsif ($found_head) {@all_tokens =  (@tokens_up_to_html, @tokens_up_to_head, $body_tag, @tokens)}
+	elsif ($found_head) {@all_tokens =  (@tokens_up_to_html, @tokens_up_to_head, $body_tag)}
 
 	# If we found <html> but no <body> and no </head> add <body> after <html>
-	elsif ($found_html) {@all_tokens = (@tokens_up_to_html, $body_tag, @tokens)}
+	elsif ($found_html) {@all_tokens = (@tokens_up_to_html, $body_tag)}
 
 	# We didn't find <body>, </head> and <html> so add <body> to the beginning
-	else {@all_tokens = ($body_tag, @tokens)}
+	else {@all_tokens = ($body_tag)}
 	
+	# If we found </body> just add the rest onto the end
+	if ($found_close_body) {push @all_tokens, @tokens_before_close_html, @tokens}
+
+	# If we found </html> but no </body> insert </body> after </html>
+	elsif ($found_close_html) {push @all_tokens, @tokens_before_close_html, $close_body_tag, @tokens}
+	
+	# We didn't find </body> or </html> add </body> to the end of the document
+	else {push @all_tokens, @tokens, $close_body_tag}
+
 	# Put them back on the parser
 	$parser->unget_token(@all_tokens);
 		
@@ -367,7 +387,7 @@ Apache::ProxyStuff - mod_perl header/footer/proxy module
 
 =head1 DESCRIPTION
 
-Apache::ProxyStuff is module for adding headers and footers to content proxied from other web servers. Rather than sandwiching the content between the header and footer it "stuffs" the header and footer into their correct places in the content -- header after the <BODY> tag and footer before the </BODY> tag. This allows you to give content living on established servers a common look and feel without making changes to the pages. (ProxyStuff will add a <BODY> tag appropriately when the document does not contain one. This allows even syntacticly incorrect pages to have a common header and footer.)
+Apache::ProxyStuff is module for adding headers and footers to content proxied from other web servers. Rather than sandwiching the content between the header and footer it "stuffs" the header and footer into their correct places in the content -- header after the <BODY> tag and footer before the </BODY> tag. This allows you to give content living on established servers a common look and feel without making changes to the pages. (ProxyStuff will add <BODY> and </BODY> tags appropriately when the document does not contain them. This allows even syntacticly incorrect pages to have a common header and footer.)
 
 ProxyStuff also allows you to add meta tags to the <HEAD> section, attributes to the <BODY> tag and manipulate links, image refs and form actions as needed.
 
